import borgmatic.actions.browse.archive

PATH_TYPE_ICONS = {
    borgmatic.actions.browse.archive.Path_type.DIRECTORY.value: '📁',
    borgmatic.actions.browse.archive.Path_type.LINK.value: '🔗',
    borgmatic.actions.browse.archive.Path_type.PIPE.value: '🚰',
    borgmatic.actions.browse.archive.Path_type.FILE.value: '📄',
}
